#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <allegro.h>

#include "sp.h"

/*#define DEBUG*/

#define SLEEP() usleep(10)
#ifdef __DJGPP__
# define LONGSLEEP() SLEEP()
#else
# define LONGSLEEP() sleep(10)
#endif

#define PAUSE_KEY KEY_P
#define QUIT_KEY KEY_Q
#define EXIT_KEY KEY_LCONTROL

static const char *program_name;

/* Coordinate pair */
typedef struct {
  unsigned char x, y;
} pos_s;

/* Movement direction */
typedef enum {
  dir_LEFT, dir_UP, dir_RIGHT, dir_DOWN, dir_NONE
} dir_e;

/* Struct for snake body parts & directions */
typedef struct {
  pos_s pos[25]; /* ring buffer */
  dir_e dir[25]; /* ring buffer */
  dir_e head_dir;
} snakepos_s;


/* Display buffer; sprites; sound samples */
static struct {
  BITMAP *board, *score;
} buffer;
static int dscale; /* 1 or 2 */
static struct {
  BITMAP *snake[7], *blob, *player;
} bmp;
static struct {
  SAMPLE *chomp, *doneit, *hiss, *munch;
} snd;

/* Snake info */
static struct {
  snakepos_s pos[7];
  unsigned char head, tail, grow; /* ring buffer indexes */
} snakes;

/* The playing area */
static unsigned char board[26][26];

/* Various playing area elements */
#define EMPTY 0
#define BLOB 1
#define SNAKEBODY 2
#define SNAKEBODY_LAST 8
#define PLAYER 9
#define SNAKETAIL 10
#define SNAKETAIL_LAST 16
#define SNAKEHEAD 17
#define SNAKEHEAD_LAST 23
#define WALL 255


/* Player position & flags */
static struct {
  pos_s pos;
  unsigned char frame;
  unsigned char snakefood, donelevel;
  unsigned char eatingsnake; /* snake being eaten (+1); 0 if none */
  struct {
    unsigned char timer, index;
  } sleep;
} player;

/* Player sleep animation sequence */
static const unsigned char sleep_anim[] = {
  4,4,4,4,4,5,
  4,4,4,4,4,6,
  4,4,4,4,4,7,
  4,4,4,4,4,7,
  8,9,8,9,8,9,8,7
};

/* Scores, game speed (units of 3/100s) */
static int score, hiscore, gamespeed;

/* Remaining on level... */
static struct {
  int snakes, blobs;
} remaining;
/* Blobs being eaten? */
static struct {
  unsigned char snake, player, snakebody;
} munch;

/* Frame update & player movement control */
static volatile struct {
  unsigned char last, this, playermove;
} frame;

/* Timer control */
static volatile struct {
  unsigned char count, ticks;
} ticker;
static int timer_running;

/* Colours, should they be needed */
#define REDS	   0
#define GREENS	  32
#define YELLOWS	  64
#define BLUES	  96
#define MAGENTAS 128
#define CYANS	 160
#define GREYS	 192
#define BLACK	   0
#define WHITE	 223


/*****
 * Misc functions
 */

inline int
quit_pressed (void)
{
  return key[QUIT_KEY] & key[EXIT_KEY];
}


/* Reset player sleep timer */
inline void
reset_sleep_timer (void)
{
  player.sleep.timer = 3;
  player.sleep.index = 0;
}


/* Run player sleep timer */
static void
do_sleep_timer (void)
{
  if (player.sleep.timer)
    player.sleep.timer--;
  if (!player.sleep.timer) {
    if (player.sleep.index >= (sizeof (sleep_anim) / sizeof (sleep_anim[0])))
      player.sleep.index = 0;
    player.frame = sleep_anim[player.sleep.index++];
    player.sleep.timer = 3;
  }
}


static void
set_player_frame (pos_s pos)
{
  if	  (pos.x < player.pos.x) player.frame = 0;
  else if (pos.x > player.pos.x) player.frame = 2;
  else if (pos.y < player.pos.y) player.frame = 1;
  else if (pos.y > player.pos.y) player.frame = 3;
}


#define IS(cell, type) ((cell) >= type && (cell) <= type##_LAST)


/*****
 * Initialise & shutdown
 */

/* Plot an arbitrary quadrilateral (two triangles) */
static void
quadrilateral (int x0, int y0, int x1, int y1,
	       int x2, int y2, int x3, int y3, int colour)
{
  triangle (screen, x0*dscale,y0*dscale, x1*dscale,y1*dscale,
	    x2*dscale,y2*dscale, colour);
  triangle (screen, x0*dscale,y0*dscale, x3*dscale,y3*dscale,
	    x2*dscale,y2*dscale, colour);
}


static void
shutdown (void)
{
  destroy_bitmap (buffer.board);
  destroy_bitmap (buffer.score);
  allegro_exit ();
}


static void
nomem (void)
{
  allegro_message ("%s: not enough memory?\n", program_name);
  exit (2);
}


/* Initialisation */
static void
init (void)
{
  DATAFILE *dat;
  char *datname = alloca (strlen (program_name) + 5);
  char *dot, *slash;

  if (!datname)
    nomem ();

  strcpy (datname, program_name);
  dot = strrchr (datname, '.');
  slash = strrchr (datname, '/');
  if (dot < slash)
    dot = 0;
  strcpy (dot ? dot : datname + strlen (datname), ".dat");

  dat = load_datafile (datname);
  if (!dat) {
    allegro_message ("%s: where's my data file?\n", program_name);
    exit (2);
  }

  /* Set up some pointers to make life easier. */
  bmp.snake[0] = (BITMAP *) dat[SNAKE0_BMP].dat;
  bmp.snake[1] = (BITMAP *) dat[SNAKE1_BMP].dat;
  bmp.snake[2] = (BITMAP *) dat[SNAKE2_BMP].dat;
  bmp.snake[3] = (BITMAP *) dat[SNAKE3_BMP].dat;
  bmp.snake[4] = (BITMAP *) dat[SNAKE4_BMP].dat;
  bmp.snake[5] = (BITMAP *) dat[SNAKE5_BMP].dat;
  bmp.snake[6] = (BITMAP *) dat[SNAKE6_BMP].dat;
  bmp.blob     = (BITMAP *) dat[BLOB_BMP].dat;
  bmp.player   = (BITMAP *) dat[PLAYER_BMP].dat;
  snd.chomp    = (SAMPLE *) dat[CHOMP_WAV].dat;
  snd.doneit   = (SAMPLE *) dat[DONEIT_WAV].dat;
  snd.hiss     = (SAMPLE *) dat[HISS_WAV].dat;
  snd.munch    = (SAMPLE *) dat[MUNCH_WAV].dat;

  /* Create the display buffer. */
  buffer.board = create_bitmap (24 * 8, 24 * 8);
  buffer.score = create_bitmap (26 * 8, 8);
  if (!buffer.board || !buffer.score)
    nomem ();

  clear (screen);
  textout (buffer.score, font, "Score          High       ", 0, 0, GREYS+26);

  set_palette ((const PALETTE *)dat[DEFAULT_PAL].dat);

  /* Draw the static bits of the display */
  stretch_blit ((BITMAP *) dat[TITLE_BMP].dat, screen,  0, 0,  220, 24,
		50*dscale, 0,  220*dscale, 24*dscale);
  quadrilateral ( 56,  24,   63,  31,  256,  31,  263,  24,  GREYS+8);
  quadrilateral ( 56,  24,   63,  31,	63, 224,   56, 231,  GREYS+16);
  quadrilateral (263, 231,  256, 224,  256,  31,  263,  24,  GREYS+12);
  quadrilateral (263, 231,  256, 224,	63, 224,   56, 231,  GREYS+20);
  line (screen,	  56*dscale,  24*dscale,   63*dscale,  31*dscale,
	dscale > 1 ? WHITE : GREYS+24);
  line (screen,  263*dscale,  24*dscale,  256*dscale,  31*dscale,
	dscale > 1 ? WHITE : GREYS+24);
  line (screen,	  56*dscale, 231*dscale,   63*dscale, 224*dscale,
	dscale > 1 ? WHITE : GREYS+24);
  line (screen,  263*dscale, 231*dscale,  256*dscale, 224*dscale,
	dscale > 1 ? WHITE : GREYS+24);
  rect (screen,	  56*dscale,  24*dscale,  263*dscale, 231*dscale,  WHITE);
  rect (screen,	  63*dscale,  31*dscale,  256*dscale, 224*dscale,  WHITE);

  text_mode (BLACK);
  srand ((int) time (0));
}


/*****
 * Timers
 */

/* Interrupt-driven ticker */
static void
frame_ticker (void)
{
  switch (--ticker.count) {
    case 0:
      ticker.count = ticker.ticks;
      frame.this++;
      break;
    case 1:
      frame.playermove = 1;
  }
}


static void
stop_ticker (void)
{
  timer_running = 0;
  remove_int (frame_ticker);
}


static void
start_ticker_lowlevel (int time, int ticks)
{
  if (timer_running)
    stop_ticker ();
  ticker.ticks = ticker.count = ticks;
  frame.last = frame.this = frame.playermove = 0;
  install_int (frame_ticker, time);
  timer_running = 1;
}


inline void
start_ticker (int milliseconds)
{
  start_ticker_lowlevel (milliseconds, 1);
}


inline void
start_game_ticker (void)
{
  start_ticker_lowlevel (30, gamespeed);
}


/*****
 * Snake/player/board plotting
 */

/* Plot snake head in buffer */
inline void
plotsnakehead (int x, int y, int snakenum)
{
  blit (bmp.snake[snakenum], buffer.board, 0,
	snakes.pos[snakenum].head_dir * 8, x*8-8, y*8-8, 8, 8);
}


/* Plot player in buffer */
static void
plotplayer (int x, int y, int frame)
{
  blit (bmp.player, buffer.board, 0, frame * 8, x*8-8, y*8-8, 8, 8);
}


/* Plot map cell in buffer (doesn't plot snake parts) */
static void
plotcell (int x, int y)
{
  switch (board[y][x]) {
    case BLOB:
      blit (bmp.blob, buffer.board, 0, 0, x*8-8, y*8-8, 8, 8);
      break;
    case PLAYER:
      plotplayer (x, y, player.frame);
      break;
    default:
      rectfill (buffer.board, x*8-8, y*8-8, x*8-1, y*8-1, BLACK);
  }
}


/* Plot whole board (exc. snakes) in buffer */
static void
plotboard (void)
{
  int x, y;
  for (y = 24; y; --y)
    for (x = 24; x; --x)
      plotcell (x, y);
}


/* Plot one snake (0..6) or all snakes (other).
 * Copes with overlap; updates board
 */
static void
plotsnake (int snakeno)
{
  int seg, snake;
  int from = ((unsigned)snakeno > 6) ? 0 : snakeno;
  int to = ((unsigned)snakeno > 6) ? 6 : snakeno;
  dir_e dir[7] = {
    dir_NONE, dir_NONE, dir_NONE, dir_NONE, dir_NONE, dir_NONE, dir_NONE
  };

  static unsigned char dirmap[][5] = {
    /* from L	U   R	D   T */
    /*L*/{  8,  7, 14,  4, 10 },
    /*U*/{  5,  9,  4, 15, 11 },
    /*R*/{ 16,  6,  8,  5, 12 },
    /*D*/{  6, 17,  7,  9, 13 }
  };

  /* wipe tail(s) from board */
  for (snake = from; snake <= to; snake++)
    for (seg = 24; seg >= 0; --seg) {
      pos_s pos = snakes.pos[snake].pos[seg];
      if (pos.x && pos.y) {
	board[pos.y][pos.x] = EMPTY;
	plotcell (pos.x, pos.y);
      }
    }

  /* draw snake body/bodies */
  seg = snakes.tail;
  do {
    for (snake = from; snake <= to; snake++) {
      pos_s pos = snakes.pos[snake].pos[seg];
      if (pos.x && pos.y) {
	dir_e newdir = snakes.pos[snake].dir[seg];
	board[pos.y][pos.x] = snake
			  + (dir[snake] == dir_NONE ? SNAKETAIL : SNAKEBODY);
	masked_blit (bmp.snake[snake], buffer.board,
		     0, dirmap[newdir][dir[snake]]*8,
		     pos.x*8-8, pos.y*8-8, 8, 8);
	dir[snake] = newdir;
      }
    }
    seg = (seg + 1) % 25;
  } while (seg != snakes.head);

  /* draw snake head(s) */
  for (snake = from; snake <= to; snake++) {
    pos_s pos = snakes.pos[snake].pos[seg];
    if (pos.x && pos.y) {
      board[pos.y][pos.x] = snake + SNAKEHEAD;
      masked_blit (bmp.snake[snake], buffer.board,
		   0, (int)snakes.pos[snake].head_dir*8,
		   pos.x*8-8, pos.y*8-8, 8, 8);
    }
  }
}


/* 'Wipe' effect */
static void
erase_board (void)
{
  int i;
  start_ticker (20);
  for (i = 0; i < 192; i += 2) {
    frame.last = frame.this;
    if (dscale > 1) {
      rectfill (screen, 128+i*2,  64, 129+i*2, 447, BLACK);
      rectfill (screen, 510-i*2,  64, 511-i*2, 447, BLACK);
      rectfill (screen, 128,  64+i*2, 511,  65+i*2, BLACK);
      rectfill (screen, 128, 446-i*2, 511, 447-i*2, BLACK);
    } else {
      vline (screen,  64+i, 32, 223, BLACK);
      vline (screen, 255-i, 32, 223, BLACK);
      hline (screen, 64,  32+i, 255, BLACK);
      hline (screen, 64, 223-i, 255, BLACK);
    }
    while (frame.last == frame.this)
      LONGSLEEP ();
  }
  stop_ticker ();
}


/* New frame - plot this one & play flagged samples */
static void
newframe (void)
{
  frame.last = frame.this;
  if (dscale > 1)
    stretch_blit (buffer.board, screen,  0, 0,  192, 192,  128, 64,  384, 384);
  else
    blit (buffer.board, screen,  0, 0,  64, 32,  192, 192);
  if (munch.snake) { /* snake eating blob */
    munch.snake = 0;
    play_sample (snd.munch, 160, 128, 1000, 0);
  }
  if (munch.player) { /* player eating blob */
    munch.player = 0;
    play_sample (snd.munch, 255, 128, 1000, 0);
  }
  if (munch.snakebody) { /* player eating snake */
    munch.snakebody = 0;
    play_sample (snd.munch, 255, 128, 500, 0);
  }
}


/*****
 * Misc game control
 */

static void
showscore (int score, int x)
{
  char digits[7];
  if ((unsigned int)score > 999999)
    strcpy (digits, "Large!");
  else
    sprintf (digits, "%06i", score);
  textout (buffer.score, font, digits, x * 8, 0, YELLOWS+26);
  if (dscale > 1)
    stretch_blit (buffer.score, screen, 0,0, 26*8,8, 112,464, 26*16,16);
  else
    blit (buffer.score, screen, 0,0, 56,232, 26*8,8);
}


/* Do sleep animation when game is paused */
static int
player_sleep (void)
{
  if (frame.last != frame.this) {
    newframe ();
    do_sleep_timer ();
    plotplayer (player.pos.x, player.pos.y, player.frame);
  }
  return quit_pressed ();
}


/* Game is paused - wait for 'P' to be released, pressed, released */
static int
pause_game (void)
{
  while (key[PAUSE_KEY])
    if (player_sleep ())
      return 1;
  while (!key[PAUSE_KEY]) {
    if (key[QUIT_KEY])
      return 2;
    if (player_sleep ())
      return 1;
  }
  while (key[PAUSE_KEY])
    if (player_sleep ())
      return 1;
  return 0;
}


/*****
 * Level init
 */

static void
initlevel (int playing)
{
  int i;
  static const pos_s start[] = {
    {11,4}, {4,5}, {19,12}, {4,19}, {19,5}, {4,12}, {19,19}, {12,20}
  };

  /* Reset snakes */
  snakes.head = 23;
  snakes.tail = 22;
  snakes.grow = 1;
#ifdef DEBUG
  remaining.blobs = 65;
#else
  remaining.blobs = 24*24;
#endif

  /* Set up & draw board */
  memset (board, WALL, 26*26);
  for (i = 24; i; --i)
    memset (&board[i][1], BLOB, 24);
  plotboard ();
  memset (snakes.pos, EMPTY, sizeof (snakes.pos));
  for (i = 6; i >= 0; i--) {
    int y;
    snakes.pos[i].pos[23] = start[i];
    snakes.pos[i].dir[23] = dir_DOWN;
    snakes.pos[i].head_dir = dir_DOWN;
    for (y = start[i].y + 2; y >= start[i].y; --y) {
      memset (&board[y][start[i].x], EMPTY, 3);
      plotcell (start[i].x    , y);
      plotcell (start[i].x + 1, y);
      plotcell (start[i].x + 2, y);
    }
    remaining.blobs -= 9;
    board[start[i].y][start[i].x] = i + SNAKEHEAD;
    plotsnakehead (start[i].x, start[i].y, i);
  }
  /* If playing, init player */
  if (playing) {
    player.pos = start[7];
    board[player.pos.y][player.pos.x] = PLAYER;
    plotplayer (player.pos.x, player.pos.y, player.frame = 4);
    remaining.blobs--;
    remaining.snakes = 7;
    reset_sleep_timer ();
    player.eatingsnake = player.snakefood = player.donelevel = 0;
  }
}


/*****
 * Movement etc.
 */

/* Move the player to a new location, erasing the old location if req. */
static void
moveplayerto (pos_s pos)
{
  if (board[player.pos.y][player.pos.x] == PLAYER) {
    board[player.pos.y][player.pos.x] = EMPTY;
    plotcell (player.pos.x, player.pos.y);
  }
  player.pos = pos;
  if (!player.snakefood
      && !IS (board[pos.y][pos.x], SNAKEBODY)
      && !IS (board[pos.y][pos.x], SNAKEHEAD)) {
    board[pos.y][pos.x] = PLAYER;
    plotplayer (pos.x, pos.y, player.frame);
  }
}


/* A blob has been eaten - count it and flag a sample */
static void
eat_blob (int is_player)
{
  if (remaining.blobs) remaining.blobs--;
  if (is_player) munch.player = 1; else munch.snake = 1;
}


/* A snake is being eaten */
static int
eat_snake (int snake)
{
  int at = snakes.tail;
  int head = (snakes.head + 1) % 25;
  snakepos_s *thissnake = &snakes.pos[snake];
  plotplayer (player.pos.x, player.pos.y, player.frame);
  start_ticker (50);
  /* Follow the snake's body */
  do {
    pos_s pos = thissnake->pos[at];
    set_player_frame (pos);
    plotsnake (snake);
    board[pos.y][pos.x] = EMPTY; /* ensure that player is drawn */
    moveplayerto (pos);
    thissnake->pos[at].y = thissnake->pos[at].x = 0;
    score += 5;
    showscore (score, 6);
    at = (at + 1) % 25;
    munch.snakebody = 1;
    /*plotcell (pos.x, pos.y);*/
    do {
      if (quit_pressed ())
	return 1;
      SLEEP ();
    } while (frame.last == frame.this);
    newframe ();
  } while (at != head);
  /* Eating the snake's head :-) */
  start_game_ticker ();
  board[player.pos.y][player.pos.x] = EMPTY;
  player.frame = 5;
  moveplayerto (player.pos);
  reset_sleep_timer ();
  memset (thissnake, 0, sizeof (snakepos_s));
  thissnake->head_dir = dir_NONE;
  if (!--remaining.snakes) {
    player.donelevel = 1;
    if (gamespeed > 3)
      gamespeed--;
  }
  return 0;
}


/* Translate dir_* to x,y change */
static pos_s
moveindir (pos_s pos, dir_e dir)
{
  switch (dir) {
    case dir_LEFT:  pos.x--; break;
    case dir_UP:    pos.y--; break;
    case dir_RIGHT: pos.x++; break;
    case dir_DOWN:  pos.y++; break;
    default: break; /* shut gcc up */
  }
  return pos;
}


/* Can the snake move in this direction? */
static int
trymovesnake (pos_s pos, dir_e dir, int snakenum)
{
  int cell;
  pos = moveindir (pos, dir);
  cell = board[pos.y][pos.x];
  return cell == EMPTY
	 || (cell == BLOB && snakenum == 0) /* red snake can eat blobs */
	 || cell == snakenum + SNAKEBODY
	 || cell == PLAYER
	 || IS (cell, SNAKETAIL);
}


/* Move the snakes */
static void
movesnakes (void)
{
  int next = (snakes.head + 1) % 25;
  int i;
  for (i = 0; i < 7; ++i)
    if (snakes.pos[i].head_dir != dir_NONE) {
      pos_s current = snakes.pos[i].pos[snakes.head];
      dir_e newdir = snakes.pos[i].head_dir;
      /* Get allowable moves info */
      int try1 = trymovesnake (current, newdir, i);
      int try2 = trymovesnake (current, (newdir+1) & 3, i);
      int try3 = trymovesnake (current, (newdir-1) & 3, i);
      int random;
      if (try1) {
	/* Can go straight on, but allow random turns */
	random = (try2 && try3 && i) ? 5 : 10;
	random = (rand () % random) - 1;
	if (random > 1)
	  random = 0;
      } else {
	/* Must turn */
	random = (rand () & 2) - 1;
      }
      if (try2 && try3) {
	/* Can turn in either direction; do so if required */
	newdir = (newdir + random) & 3;
      } else if (try2) {
	/* Maybe ahead or right */
	if (rand () & 1 || !try1) newdir = (newdir + 1) & 3;
      } else if (try3) {
	/* Maybe ahead or left */
	if (rand () & 1 || !try1) newdir = (newdir - 1) & 3;
      } else {
	/* U-turn */
	if (!try1) newdir ^= 2;
      }
      /* New head position */
      current = moveindir (current, newdir);
      if (board[current.y][current.x] == BLOB)
	eat_blob (0);
      board[current.y][current.x] = SNAKEHEAD + i;
      snakes.pos[i].pos[next] = current;
      snakes.pos[i].dir[next] = newdir;
      snakes.pos[i].dir[snakes.head] = newdir;
      snakes.pos[i].head_dir = newdir;
    }
  /* Update the head & tail positions
   * (ring buffer, so adjust head/tail pointers) */
  snakes.head = next;
  if (snakes.grow < 23)
    snakes.grow++;
  else
    snakes.tail = (snakes.tail + 1) % 25;
  plotsnake (-1);
}


/* Move the player in the given direction if possible. Returns 1 if moved. */
static int
tryplayermove (int x, int y)
{
  int cell;
  pos_s pos;
  pos.x = x + player.pos.x;
  pos.y = y + player.pos.y;
  cell = board[pos.y][pos.x];
  if (cell == WALL || IS (cell, SNAKEBODY) || cell == PLAYER)
    return 0; /* cell occupied by snake body (or not moving) - can't move */
  if (IS (cell, SNAKEHEAD))
    player.snakefood = 1; /* whoops, we've run into a snake's head */
  if (!player.snakefood) {
    if (pos.x || pos.y)
      reset_sleep_timer ();
    if (!remaining.blobs && IS (cell, SNAKETAIL))
      /* all blobs eaten, on a snakes's tail -> prepare to eat the snake */
      player.eatingsnake = cell - SNAKETAIL + 1;
    if (cell == 1) {
      /* moved onto a blob - eat it */
      eat_blob (1);
      score += 10;
      showscore (score, 6);
    }
    set_player_frame (pos);
  }
  moveplayerto (pos);
  return 1;
}


/* Player movement control. Returns 1 if moved. */
static int
moveplayer (void)
{
  int x = (key[KEY_X] != 0 || key[KEY_RIGHT] != 0)
	- (key[KEY_Z] != 0 || key[KEY_LEFT] != 0);
  int y = (key[KEY_SLASH] != 0 || key[KEY_DOWN] != 0)
	- (key[KEY_QUOTE] != 0 || key[KEY_UP] != 0);
  static int playerdir = 0; /* for diagonal movement */

  playerdir = !playerdir;
  if (x || y) {
    if (playerdir)
      return tryplayermove (0, y) || tryplayermove (x, 0);
    return tryplayermove (x, 0) || tryplayermove (0, y);
  }
  return tryplayermove (x, y);
}


/*****
 * Demo & game loops
 */

typedef struct {
  const char *text;
  const char *colour;
  int frames;
} demotext_s;


/* Demo mode */
static int
demo (void)
{
  static const demotext_s demotext[] = {
    /*...|...|...|...|...|...|..*/
    { "Press SPACE to play.",
      "", 10 },
    { "A RETROSPEC production.",
      "^>", 5 },
    { "Written by Darren Salt.",
      "", 5},
    { "Press SPACE to play.",
      "", 10 },
    { "Eat the 'eggs' then",
      "^^^^ZZZZ\\\\\\\\\\\\XXXXX", 5 },
    { "eat the snakes.",
      "^^^^ZZZZ\\\\\\\\\\\\X", 5},
    { "Keys: ZX'/ or cursors.",
      "", 5},
    { "Press P to pause & resume.",
      "<<<<<<><<<<<<<<<<<<<<<<<<<", 5},
    { "Ctrl-Q to exit.",
      "", 5},
    { 0, 0, 0 }
  };

  BITMAP *demo_bmp;
  int ret = -1;
  int textind = -1, demoind = 0;
  int framecount = 1;
  int demotext_lines = 0;

  while (demotext[++demotext_lines].frames)
    ;

  /* oh look, a vertical scrolly */
  demo_bmp = create_bitmap (26 * 8, demotext_lines * 12 + 16);
  if (!demo_bmp)
    nomem ();

  clear (demo_bmp);
  for (textind = 0; textind < demotext_lines; textind++) {
    int x, i = 0;
    char tmp[2] = { 0, 0 };
    framecount = demotext[textind].frames;
    x = 26*4 - 8 - strlen (demotext[textind].text) * 4;
    while (demotext[textind].text[i]) {
      tmp[0] = demotext[textind].text[i];
      textout (demo_bmp, font, tmp, (x += 8), (textind * 12 + 8),
	       demotext[textind].colour[i++]);
    }
  }
  /* copy top line to bottom */
  blit (demo_bmp, demo_bmp, 0,8, 0,(textind * 12 + 8), 26*8,8);

  textind = 0;

  initlevel (0);
  start_ticker (300);
  /* demo mode loop */
  do {
    newframe ();
    if ((demoind + 4) % 12 || !--framecount) {
      if (!framecount) {
	textind = (textind + 1) % demotext_lines;
	framecount = demotext[textind].frames + 12;
      }
      if (++demoind == demotext_lines * 12 + 8) demoind = 8;
      if (dscale > 1)
	stretch_blit (demo_bmp, screen, 0,demoind, 26*8,8,
		      112,464, 26*16,16);
      else
	blit (demo_bmp, screen, 0,demoind, 56,232, 26*8,8);
    }
    movesnakes ();
    do {
      if (key[KEY_SPACE])
	ret = 0;
      else if (quit_pressed ())
	ret = 1;
      SLEEP ();
    } while (frame.last == frame.this && ret < 0);
  } while (ret < 0);
  stop_ticker ();
  destroy_bitmap (demo_bmp);
  return ret;
}


/* Main game loop */
static int
game (void)
{
  score = 0;
  gamespeed = 10;
  erase_board ();
  showscore (0, 6);
  showscore (hiscore, 20);
  /* One game */
  do {
    initlevel (1);
    play_sample (snd.hiss, 127, 128, 1000, 1);
    start_game_ticker ();
    /* One level */
    do {
      newframe ();
      if (player.eatingsnake) {
	if (eat_snake (player.eatingsnake - 1))
	  return 1;
	player.eatingsnake = 0;
      }
      movesnakes ();
      do {
	if (quit_pressed ())
	  return 1;
	if (key[PAUSE_KEY]) {
	  switch (pause_game ()) {
	    case 1: return 1;
	    case 2: player.snakefood = 1; frame.playermove = 1;
	      if (board[player.pos.y][player.pos.x] == PLAYER) {
		board[player.pos.y][player.pos.x] = EMPTY;
		plotcell (player.pos.x, player.pos.y);
	      }
	  }
	}
	if (frame.playermove) {
	  /* move the player a short while after frame update */
	  frame.playermove = 0;
	  if (!moveplayer ()) {
	    /* if not moved for a few frames, sleep */
	    do_sleep_timer ();
	    plotplayer (player.pos.x, player.pos.y, player.frame);
	  }
	}
	SLEEP ();
      } while (frame.last == frame.this);
    } while (!player.snakefood && !player.donelevel);
    /* Player has eaten all of the snakes or was eaten by a snakes. */
    stop_sample (snd.hiss);
    if (player.donelevel)
      play_sample (snd.doneit, 255, 128, 1000, 0);
    else if (player.snakefood) {
      play_sample (snd.chomp, 255, 128, 500, 0);
      if (score > hiscore) {
	hiscore = score;
	textout (screen, font, "beats", 168, 232, CYANS+20);
      }
    }
    /* Wait for player to press Space while animating remaining snakes. */
    do {
      newframe ();
      movesnakes ();
      do {
	if (quit_pressed ())
	  return 1;
	SLEEP ();
      } while (frame.last == frame.this && !key[KEY_SPACE]);
    } while (!key[KEY_SPACE]);
    /* Wait for player to release Space... */
    do {
      newframe ();
      movesnakes ();
      do {
	if (quit_pressed ())
	  return 1;
	SLEEP ();
      } while (frame.last == frame.this && key[KEY_SPACE]);
    } while (key[KEY_SPACE]);
  } while (!player.snakefood);
  return 0;
}


/*****
 * Main program
 */


int
main (int argc, char **argv)
{
  program_name = argv[0];
  dscale = (argc > 1 && !strcmp (argv[1], "-2")) ? 2 : 1;

  allegro_init ();
  atexit (shutdown);
  install_keyboard ();
  install_timer ();
  reserve_voices (4, 0);
  install_sound (DIGI_AUTODETECT, MIDI_NONE, 0);

  set_color_depth (8);
  set_color_conversion (COLORCONV_NONE);
  while (set_gfx_mode (GFX_AUTODETECT, 320 * dscale, 240 * dscale, 320, 240) < 0)
  {
    allegro_message ("%s: %s %ix mode isn't available: %s\n", program_name,
		     (dscale > 1 ? "warning: " : ""), dscale, allegro_error);
    if (!--dscale)
      exit (2);
  }

  init ();

  if (!demo ())
    while (!game ());

  return 0;
}
END_OF_MAIN ();
